#output {
    padding: 20px;
    padding-left: 40px;
    box-shadow: 2px 0 5px rgba(0, 0, 0, 0.1);
    flex-grow: 1;
    max-width: 100%;
}

.pagination {
    text-align: center;
    margin: 20px 0;
}

.pagination-text {
    display: inline-block;
    padding: 8px 12px;
    margin: 4px;
    color: #555;
}

.pagination-link {
    display: inline-block;
    padding: 8px 12px;
    margin: 4px;
    text-decoration: none;
    color: #007bff;
    border: 1px solid #007bff;
    border-radius: 4px;
    transition: background-color 0.3s, color 0.3s;
}

.pagination-link:hover {
    background-color: #007bff;
    color: white;
}

.pagination-link.active {
    background-color: #007bff;
    color: white;
    pointer-events: none;
}

.pagination-link-edge {
    text-decoration: underline;
}

.gallery-container {
    gap: 16px;
}

.gallery-container-grid {
    display: grid;
    grid-template-columns: repeat(auto-fit, minmax(360px, 1fr));
}

.gallery-container-list {
    display: grid;
}

.gallery-container-hide-images .embed-image {
    display: none;
}

.gallery-container-hide-images .embed-thumbnail {
    display: none;
}

.gallery-container-clamp-description .embed-description {
    display: -webkit-box;
    -webkit-box-orient: vertical;
    -webkit-line-clamp: 6;
    overflow: hidden;
    overflow: clamp;
    text-overflow: ellipsis;
    line-clamp: 6;
}

.embed-header {
    display: flex;
    flex-direction: row;
    flex-wrap: wrap;
}

.embed-body {
    display: flex;
    flex-direction: row;
    flex-wrap: wrap;
}

.gallery-card {
    background-color: #323339;
    color: rgb(243, 243, 244);
    border-radius: 3px;
    padding: 12px;
    box-shadow: 0 2px 4px rgba(0, 0, 0, 0.1);
    display: flex;
    flex-direction: column;
}

.gallery-card-timestamp {
    font-size: 12px;
    font-style: italic;
    color: #b9bbbe;
    padding: 4px;
    padding-top: 8px;
    margin-top: auto;
    margin-left: auto;
}

.embed-container {
    background-color: #393a41;
    box-shadow: 0 0 1px #9e9e9ecc;
    border-radius: 2px;
    padding: 12px;
    margin-top: 8px;
    position: relative;
    display: flex;
    flex-direction: column;
}

.embed-container a {
    color: #79acf0;
}

.embed-container pre {
    background-color: #555555;
    color: #dceffd;
    padding: 15px;
    border-radius: 5px;
    font-family: monospace;
    line-height: 1.5em;
    white-space: pre-wrap;
    overflow: auto;
}

.embed-container code {
    background-color: #555555;
    color: #dceffd;
    padding: 2px 2px;
    border-radius: 1px;
    font-family: monospace;
}

.embed-container blockquote {
    background-color: #23272a;
    color: #d3d3d3;
    border-left: 4px solid #7289da;
    padding: 10px;
    margin: 10px 0;
    border-radius: 5px;
    font-family: 'Arial', sans-serif;
}

.embed-author {
    display: flex;
    align-items: center;
    padding-bottom: 6px;
}

.embed-icon-container {
    display: flex;
}

.embed-icon {
    border-radius: 50%;
    width: 25px;
    height: 25px;
    margin-right: 5px;
    object-fit: cover;
}

.embed-icon-preview {
    display: none;
    position: absolute;
    z-index: 10;
    max-width: 400px;
}

.embed-icon-container:hover .embed-icon-preview {
    display: block;
}

.embed-author-name {
    font-weight: bold;
    text-decoration: none;
    color: #f3f3f3;
    margin-right: 8px;
}

.embed-title {
    font-size: 16px;
    font-weight: bold;
    color: #79acf0;
    text-decoration: none;
    display: block;
    max-width: 100%;
    overflow-wrap: break-word;
}

.embed-description {
    margin: 4px 0;
    max-width: 100%;
    overflow-wrap: break-word;
    flex: 3 1 60%;
}

.embed-fields {
    display: flex;
    flex-wrap: wrap;
    margin: 10px 0;
}

.embed-field {
    flex: auto;
    text-align: left;
}

.embed-field-name {
    color: #e9ecf1;
    font-weight: bold;
}

.embed-field-value {
    color: #b9bbbe;
    margin-top: 2px;
}

.embed-image,
.embed-thumbnail {
    max-width: 400px;
    width: 100%;
    height: fit-content;
    min-height: 2em;
    flex: 1 3 40%;
    border-radius: 4px;
    margin: 4px;
}

.embed-thumbnail {
    max-width: 200px;
}

.embed-footer-content {
    display: flex;
    align-items: center;
    margin-top: 10px;
    font-size: 12px;
    color: #b9bbbe;
}

.embed-footer-text {
    margin-left: 5px;
}

.embed-timestamp {
    font-size: 12px;
    color: #b9bbbe;
    margin-top: 5px;
    font-style: italic;
}

.fullsize-image-container {
    background-color: rgba(45, 45, 45, 0.9);
    display: flex;
    flex: 1 1 80%;
    justify-content: center;
    max-width: 80%;
    max-height: 80%;
    padding: 20px;
}

.fullsize-image {
    max-width: 100%;
    max-height: 100%;
}

.controls {
    display: flex;
    flex-direction: column;
    padding: 12px;
    margin-top: auto;
    margin-left: auto;
    margin-right: auto;
    margin-bottom: 20px;
    position: sticky;
    background-color: #e0e0e0;
    border-radius: 4px;
    cursor: pointer;
    transition: background-color 0.3s;
}

.controls-group {
    display: flex;
    justify-content: center;
    padding: 8px;
    margin-top: 1px;
    margin-bottom: 1px;
    background-color: #f0f0f0;
    border-radius: 4px;
    box-shadow: 0 2px 5px rgba(0, 0, 0, 0.1);
}

.controls-button {
    display: flex;
    align-items: center;
    justify-content: center;
    font-size: 24px;
    color: #007bff;
    background-color: #e0e0e0;
    border: none;
    border-radius: 4px;
    padding: 4px 8px;
    margin: 2px;
    cursor: pointer;
    transition: background-color 0.3s, color 0.3s;
}

.controls-button:hover {
    background-color: #d0d0d0;
}

.controls-button:active {
    background-color: #c0c0c0;
}

.controls-button:disabled {
    cursor: default;
}

.highlight {
    transition: background-color 2s ease-out;
}

.bg-highlight {
    background-color: #5b5a5a;
}
